import sys

from fpy.composable.function import func


@func
def trace(txt, val):
    print(txt, val, file=sys.stderr)
    return val


def showio(fn):
    def res(*args, **kwargs):
        print(f"{fn.__name__} input: {args, kwargs}", file=sys.stderr)
        ret = fn(*args, **kwargs)
        print(f"{fn.__name__} output: {ret}", file=sys.stderr)
        return ret

    return res



from fpy.composable.function import func, SignatureMismatchError, NotEnoughArgsError
from fpy.composable.collections import (
    Seq,
    Map,
    transN,
    setN,
    getN,
    get0,
    of_,
    is_,
    and_,
    or_,
    to,
    apply,
    mapN,
    fwd_,
)
from fpy.data.forgetful import Under
from fpy.utils.placeholder import __

import unittest


def add1(x):
    return x + 1


def mul2(x):
    return x * 2


def pos(x):
    return x > 0


def odd(x):
    return x % 2 != 0


def add(a, b):
    return a + b


class TestCompsable(unittest.TestCase):
    def testFunc(self):
        f1 = func(add1)
        f2 = func(mul2)
        self.assertEqual(4, (f1 ^ f2)(1))
        self.assertEqual(3, (f2 ^ f1)(1))
        self.assertEqual(fwd_(1) | f1 ^ f2, Under(4))
        self.assertEqual(fwd_(1) | f2 ^ f1, Under(3))

    def testSeq(self):
        lst = [1, 2, 3, 4, 5]
        fn = Seq(lst)
        self.assertEqual(1, fn(0))
        self.assertEqual(2, fn(1))
        self.assertEqual(3, fn(2))
        self.assertEqual(4, fn(3))
        self.assertEqual(5, fn(4))
        self.assertEqual(Under(1), fwd_(0) | fn)
        self.assertEqual(Under(2), fwd_(1) | fn)
        self.assertEqual(Under(3), fwd_(2) | fn)
        self.assertEqual(Under(4), fwd_(3) | fn)
        self.assertEqual(Under(5), fwd_(4) | fn)

    def testMap(self):
        mp = {"a": 1, "b": 2, "c": 3}
        fn = Map(mp)
        self.assertEqual(1, fn("a"))
        self.assertEqual(2, fn("b"))
        self.assertEqual(3, fn("c"))
        self.assertEqual(Under(1), fwd_("a") | fn)
        self.assertEqual(Under(2), fwd_("b") | fn)
        self.assertEqual(Under(3), fwd_("c") | fn)

    def testItN(self):
        lst = [1, 2, 3]
        self.assertListEqual([2, 2, 3], transN(0, add1, lst))
        self.assertListEqual([2, 2, 3], setN(0, 2, lst))
        self.assertEqual(2, getN(1, lst))

    def testOf(self):
        self.assertTrue(of_(1, 2, 3, 4)(2))
        self.assertFalse(of_(1, 2, 3, 4)(5))

    def testIs(self):
        self.assertTrue(is_(int)(1))
        self.assertFalse(is_(str)(1))

    def testAnd(self):
        self.assertTrue(and_(odd, pos)(1))
        self.assertFalse(and_(odd, pos)(2))

    def testOr(self):
        self.assertTrue(or_(odd, pos)(1))
        self.assertTrue(or_(odd, pos)(2))

    def testTo(self):
        fn = to(int)
        self.assertEqual(1, fn("1"))

    def testApply(self):
        fn = apply(add1)
        self.assertEqual(2, fn((1,)))

    def testMapN(self):
        f1 = mapN(1, add1)
        self.assertListEqual([2, 3, 4], f1([1, 2, 3]))
        f2 = mapN(1)(add1)
        self.assertListEqual([2, 3, 4], f2([1, 2, 3]))

    def testPlaceholder(self):
        class T:
            def __init__(self, a):
                self.a = a

        a = T(1)
        f = __.a == 1
        self.assertTrue(f(a))

        b = T([1,2,3])
        g = __.a ^ get0 == 1
        self.assertTrue(g(b))



from __future__ import annotations

import inspect
import collections.abc as cabc

from fpy.composable.composable import Composable
from fpy.composable.transparent import Transparent

from typing import Callable, TypeVar, Generic, Union, Any, Optional

import sys
import traceback

class SignatureMismatchError(Exception):
    def __init__(self, e):
        self.e = e


class NotEnoughArgsError(Exception):
    def __init__(self, expect, got):
        self.expect = expect
        self.got = got

R = TypeVar("R")

class func(Composable, Transparent, Generic[R]):
    fn: Optional[Callable[[Any], R]] = None

    def __init__(self, f, *args, **kwargs):
        if isinstance(f, func):
            self.fn = f.fn
            self.args = (*f.args, *args)
            self.kwargs = {**f.kwargs, **kwargs}
            self.sig = f.sig
        elif isinstance(f, cabc.Callable):
            self.fn = f
            self.args = args
            self.kwargs = kwargs
            self.sig = inspect.signature(f)
        else:
            raise TypeError(f"{f} is not callable")

    def __underlying__(self):
        return self.fn

    def __repr__(self):
        return repr(self.fn)

    def __call__(self, *args, **kwargs) -> Union[R, func[R]]:
        _args = (*self.args, *args)
        _kwargs = {**self.kwargs, **kwargs}
        try:
            self.sig.bind(*_args, **_kwargs)
            return self.fn(*_args, **_kwargs)
        except TypeError as e:
            # print(f"TypeError: {e}", file=sys.stderr)
            # tb = sys.exc_info()[2]
            # print(f"{tb.tb_frame.f_code.co_filename = }, {tb.tb_lineno = }")
            # traceback.print_tb(tb)
            try:
                self.sig.bind_partial(*_args, **_kwargs)
                return func(self, *args, **kwargs)
            except TypeError as e:
                raise SignatureMismatchError(e)
        except NotEnoughArgsError:
            return func(self, *args, **kwargs)

    def __compose__(self, other):
        if isinstance(other, func):
            return self.__class__(lambda *args, **kwargs: other(self(*args, **kwargs)))
        if isinstance(other, cabc.Callable):
            fn = func(other)
            return self.__class__(lambda *args, **kwargs: fn(self(*args, **kwargs)))
        raise TypeError(f"Cannot compose function with {type(other)}")



from __future__ import annotations

from dataclasses import dataclass
from typing import TypeVar, Generic, Callable

from fpy.control.monad import _Monad
from fpy.control.natural_transform import NTrans, _NTrans


T = TypeVar("T")
F = TypeVar("F")
G = TypeVar("G")
B = TypeVar("B")


@dataclass
class Under(_Monad, Generic[T]):
    v: T

    def under(self):
        return self.v

    def __fmap__(self, f: Callable[[T], B]) -> Under[B]:
        return Under(f(self.v))

    def __ntrans__(self, t: Callable[[T], G[B]]) -> G[B]:
        return t(self.val)

    def __bind__(self, b: Callable[[T], B]) -> Under[B]:
        return self.__fmap__(b)


forget: _NTrans[F, B, Under, T] = NTrans(Under)



from __future__ import annotations

from fpy.composable.function import func
from fpy.control.monad import _Monad
from fpy.control.natural_transform import NTrans, _NTrans
from abc import ABCMeta, abstractmethod, abstractclassmethod
import collections.abc as cabc
from dataclasses import dataclass
from typing import TypeVar, Generic, Callable, List, Tuple


T = TypeVar("T")
R = TypeVar("R")
L = TypeVar("L")
G = TypeVar("G")


class Either(_Monad, Generic[L, R]):
    @abstractmethod
    def __bool__(self):
        return NotImplemented

    def __bind__(self, b):
        raise NotImplementedError


@dataclass
class Left(Either[L, R]):
    v: L

    def __bool__(self):
        return False

    def __fmap__(self, f: Callable[[L], T]) -> Left[T]:
        return self

    def __bind__(self, _) -> Left[L, R]:
        return self

    def __ntrans__(self, t: _NTrans[Either, [L, R], G, T]) -> G[T]:
        return t(self.v)


@dataclass
class Right(Either[L, R]):
    v: R

    def __bool__(self):
        return True

    def __fmap__(self, f: Callable[[R], T]) -> Right[T]:
        return Right(f(self.v))

    def __bind__(self, b: Callable[[R], Either[T, G]]) -> Either[T, G]:
        return b(self.v)

    def __ntrans__(self, t: _NTrans[Either, [L, R], G, T]) -> G[T]:
        return t(self.v)

    def __enter__(self):
        return self.v


def isLeft(e: Either[L, R]) -> bool:
    return isinstance(e, Left)


def isRight(e: Either[L, R]) -> bool:
    return isinstance(e, Right)


def fromLeft(d: L, e: Either[L, R]) -> L:
    return d if not isLeft(e) else e.v


def fromRight(d: R, e: Either[L, R]) -> R:
    return d if not isRight(e) else e.v


@func
def either(fl: Callable[[L], T], fr: Callable[[R], T], e: Either[L, R]) -> T:
    if isLeft(e):
        return fl(fromLeft(None, e))
    return fr(fromRight(None, e))


def lefts(l: cabc.Sequence[Either[L, R]]) -> List[L]:
    return [fromLeft(None, v) for v in l if isLeft(v)]


def rights(l: cabc.Sequence[Either[L, R]]) -> List[R]:
    return [fromRight(None, v) for v in l if isRight(v)]


r2l: _NTrans[Right, T, Left, T] = NTrans(Left)
l2r: _NTrans[Left, T, Right, T] = NTrans(Right)


def partitionEithers(lst: cabc.Sequence[Either[L, R]]) -> Tuple[List[L], List[R]]:
    l = []
    r = []
    for e in lst:
        if isLeft(e):
            l.append(fromLeft(None, e))
        else:
            r.append(fromRight(None, e))
    return l, r



from fpy.data.function import id_, const, flip, fix, on, constN, uncurryN

import unittest


def catInt(a: str, b: int) -> str:
    return a + str(b)


def add(a: int, b: int) -> int:
    return a + b


class TestFunction(unittest.TestCase):
    def testId(self):
        v = id_(1)
        self.assertEqual(1, v)

    def testConst(self):
        f = const(1)
        self.assertEqual(1, f(123))

    def testFlip(self):
        a = 123
        b = "hello"
        f = flip(catInt)
        self.assertEqual("hello123", f(a, b))

    def testOn(self):
        a = "1"
        b = "2"
        f = on(add, int)
        self.assertEqual(3, f(a, b))

    def testFix(self):
        _fib = lambda r, n: 1 if n <= 1 else n * r(n - 1)
        fib = fix(_fib)
        self.assertEqual(120, fib(5))

    def testConstN(self):
        f = constN(3, 42)
        self.assertEqual(42, f(1)(2)(3))

    def testUncurryN(self):
        f = constN(3, 42)
        uf = uncurryN(3, f)
        self.assertEqual(42, uf(1, 2, 3))

    def testUncurryNErr(self):
        f = constN(3, 42)
        uf = uncurryN(3, f)
        with self.assertRaises(TypeError):
            uf(1, 2)



from abc import ABCMeta, abstractmethod

from fpy.control.applicative import _Applicative
from typing import TypeVar, Generic, Callable


T = TypeVar("T")
R = TypeVar("R")


class _Monad(_Applicative[T], Generic[T]):
    @classmethod
    def ret(cls, val):
        return cls(val)

    def __bind__(self, b):
        raise NotImplementedError

    def __rshift__(self, b):
        return self.__bind__(b)

    def __enter__(self):
        raise DoFail(self)

    def __exit__(self, et, ev, tb):
        pass


class DoFail(Exception):
    def __init__(self, m):
        super()
        self.m = m

def do(fn):
    def res(*args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except DoFail as f:
            return f.m
    return res



from fpy.parsec.parsec import (
    parser,
    one,
    neg,
    pmaybe,
    many,
    many1,
    ptrans,
    peek,
    skip,
    pseq,
    inv,
)
from fpy.data.either import fromRight

import unittest

toks = [1, 2, 3]

even = lambda x: x % 2 == 0
odd = lambda x: x % 2 != 0


class TestParsec(unittest.TestCase):
    def testParseOne(self):
        p = one(odd)
        self.assertTrue(p(toks))
        p = one(even)
        self.assertFalse(p(toks))

    def testParseNeg(self):
        p = neg(odd)
        self.assertFalse(p(toks))
        p = neg(even)
        self.assertTrue(p(toks))

    def testParseCat(self):
        p1 = one(odd)
        p2 = one(even)
        p = p1 + p2
        self.assertTrue(p(toks))

    def testParseLR(self):
        p1 = one(odd)
        p2 = one(even)
        pl = p1 << p2
        pr = p1 >> p2

        self.assertTrue(pl(toks))
        self.assertTrue(pr(toks))

    def testParseChoice(self):
        p1 = one(odd)
        p2 = one(even)
        p = p1 | p2
        self.assertTrue((p + p)(toks))

    def testMany(self):
        p1 = one(odd)
        p2 = one(even)
        p = many1(p1)
        self.assertTrue(p(toks))
        p = many(p2)
        self.assertTrue(p(toks))

    def testSkip(self):
        p1 = one(odd)
        p2 = one(even)
        p = skip(p1) + p2
        self.assertTrue(p(toks))

    def testPSeq(self):
        seq = [1, 2, 3]
        p = pseq(seq)
        self.assertTrue(p(toks))
        self.assertFalse(pseq([2, 3, 4])(toks))

    def testPeek(self):
        p1 = one(odd)
        p = peek(p1)
        res = p(toks)
        self.assertTrue(res)
        head, rest = fromRight(None, res)
        self.assertEqual(head, [1])
        self.assertIs(toks, rest)

    def testTimeN(self):
        p1 = one(odd)
        p2 = one(even)
        p = p1 | p2
        self.assertTrue((p * 2)(toks))

    def testInv(self):
        p1 = pseq("abc")
        p2 = inv(p1)
        p3 = pseq("cba")
        self.assertFalse(p2("abc"))
        self.assertTrue((p2 >> p3)("cba"))



import bytecode as bc
import fpy.experimental.pattern.core as pat_core

from fpy.data.function import const
from fpy.composable.collections import is_, and_, or_, get1, mp1, trans0, get0, to
from fpy.composable.function import func
from fpy.data.forgetful import forget
from fpy.data.maybe import Just, Nothing, isJust, fromJust
from fpy.data.either import either, fromRight
from fpy.experimental.do import do
from fpy.parsec.parsec import one, many, neg, skip, many1, ptrans, s2c
from fpy.utils.placeholder import __

from fpy.debug.debug import trace

from dataclasses import dataclass
from typing import List

import sys

import dis

isInstr = is_(bc.Instr)
popTop = and_(isInstr, __.name == "POP_TOP")
none = and_(isInstr, and_(__.name == "LOAD_CONST", __.arg == None))
ret = and_(isInstr, __.name == "RETURN_VALUE")
ending = one(popTop) >> one(none) >> one(ret)
mkConstMap = and_(isInstr, __.name == "BUILD_CONST_KEY_MAP")
mkMap = and_(isInstr, __.name == "BUILD_MAP")
isArg = and_(isInstr, __.name == "LOAD_FAST")
isLoadGlobal = and_(isInstr, __.name == "LOAD_GLOBAL")
isVarName = lambda x: x != "_" and x.startswith("_")
if sys.version_info.major == 3 and sys.version_info.minor >= 11:
    isVar = and_(isLoadGlobal, __.arg ^ get1 ^ isVarName)
    callInst = one(and_(isInstr, __.name == "PRECALL")) >> one(
        and_(isInstr, __.name == "CALL")
    )
    isWildcard = and_(isLoadGlobal, __.arg ^ get1 == "_")
else:
    isWildcard = and_(isLoadGlobal, __.arg == "_")
    isVar = and_(isLoadGlobal, __.arg ^ isVarName)
    callInst = one(and_(isInstr, __.name == "CALL_FUNCTION"))

storeFast = and_(isInstr, __.name == "STORE_FAST")
unpack = and_(isInstr, __.name == "UNPACK_SEQUENCE")

isVarargFn = lambda fn: 1 == ((fn.__code__.co_flags >> 2) & 1)

rightToMaybe = either(const(Nothing()), Just)

@dataclass
class CaseHead:
    args: List[str]


@dataclass
class CaseBody:
    insts: List[bc.Instr]


@dataclass
class StoreName:
    instrs: List[bc.Instr]


@dataclass
class Case:
    head: CaseHead
    body: CaseBody
    store: StoreName


def exprToLambda(b, place, filename, args, fv, v=None):
    varMap = {arg: pat_core._fresh(arg) for arg in args}
    mod_b = []
    free_vars = []
    # print(f"{b = }")
    for instr in b:
        if isArg(instr):
            name = instr.arg
            if name in varMap:
                instr.arg = varMap[name]
            else:
                instr = bc.Instr("LOAD_DEREF", bc.FreeVar(name))
                free_vars.append(name)
        if isVar(instr):
            if sys.version_info.major == 3 and sys.version_info.minor >= 11:
                name = instr.arg[1]
            else:
                name = instr.arg
            assert name in v, f"Variable: {name} is not bound"
            instr = bc.Instr("LOAD_FAST", varMap[v[name]])
        if isLoadGlobal(instr):
            name = instr.arg[1]
            if name in varMap:
                instr = bc.Instr("LOAD_FAST", varMap[name])
        mod_b.append(instr)
    mod_b.append(bc.Instr("RETURN_VALUE"))
    if free_vars and sys.version_info.major == 3 and sys.version_info.minor >= 11:
        mod_b.insert(0, bc.Instr("COPY_FREE_VARS", len(free_vars)))
    # print(f"{mod_b = }")
    lm = bc.Bytecode(mod_b)
    lm.freevars.extend(fv)
    lm.freevars.extend(free_vars)
    lm.argcount = len(varMap)
    lm.argnames.extend(list(varMap.values()))
    lm.name = pat_core._fresh(place, "exprlambda")
    lm.filename = filename
    lm.flags = lm.flags | 16
    lm.update_flags()
    co = lm.to_code()
    # print("=" * 20)
    # dis.dis(co)
    # dis.show_code(co)
    # print("=" * 20)
    return [
        *[bc.Instr("LOAD_CLOSURE", bc.CellVar(f)) for f in free_vars],
        *([bc.Instr("BUILD_TUPLE", len(free_vars))] if free_vars else []),
        bc.Instr("LOAD_CONST", co),
        *([] if sys.version_info.major == 3 and sys.version_info.minor >= 11 else [bc.Instr("LOAD_CONST", lm.name)]),
        bc.Instr("MAKE_FUNCTION", 0x08 if free_vars else 0),
    ], free_vars


def generateDefault(place, filename, args):
    varMap = {arg: pat_core._fresh(arg) for arg in args}
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        lm = bc.Bytecode(
            [
                bc.Instr("PUSH_NULL"),
                bc.Instr("LOAD_CONST", pat_core.NonExhaustivePatternError),
            ]
        )
    else:
        lm = bc.Bytecode([bc.Instr("LOAD_CONST", pat_core.NonExhaustivePatternError)])
    lm.append(bc.Instr("LOAD_CONST", f"Non Exhaustive Pattern Matching: {place}"))
    for k, v in varMap.items():
        lm.append(bc.Instr("LOAD_CONST", f"\n{k}: "))
        lm.append(bc.Instr("LOAD_FAST", v))
        lm.append(bc.Instr("FORMAT_VALUE", 0x02))
    lm.append(bc.Instr("BUILD_STRING", 1 + 2 * len(varMap)))
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        lm.append(bc.Instr("PRECALL", 1))
        lm.append(bc.Instr("CALL", 1))
    else:
        lm.append(bc.Instr("CALL_FUNCTION", 1))
    lm.append(bc.Instr("RAISE_VARARGS", 1))
    lm.argcount = len(varMap)
    lm.argnames.extend(list(varMap.values()))
    lm.name = pat_core._fresh(place, "defaultcase")
    lm.filename = filename
    lm.flags = lm.flags | 16
    lm.update_flags()
    co = lm.to_code()
    return [
        bc.Instr("LOAD_CONST", co),
        *([] if sys.version_info.major == 3 and sys.version_info.minor >= 11 else [bc.Instr("LOAD_CONST", lm.name)]),
        bc.Instr("MAKE_FUNCTION", 0),
    ]


def partitionInst(insts, n):
    if not insts:
        return [], []
    if n == 0:
        return [], insts
    head = insts[-1]
    # print(f"{head = }")
    # print(f"{n = }")
    pre, post = head.pre_and_post_stack_effect()
    # print(f"{pre = }")
    # print(f"{post= }")
    if pre > 0:
        if pre == n:
            return [head], insts[:-1]
        if pre < n:
            nxt, rst = partitionInst(insts[:-1], n - pre)
            return nxt + [head], rst
    if pre == 0:
        nxt, rst = partitionInst(insts[:-1], n - post)
        return nxt + [head], rst
    pre = abs(pre)
    nxt, rst = partitionInst(insts[:-1], pre)
    if post < n:
        head = nxt + [head]
        nxt, rst = partitionInst(rst, n - post)
        return nxt + head, rst
    # if post == n:
    return nxt + [head], rst


def transConstMap(case_inst: Case, fn_name, filename, args, fv):
    body = case_inst.body.insts[:-1]
    mk = case_inst.body.insts[-1]
    keynames = body[-1]
    exprs = []
    rest = body[:-1]
    args = case_inst.head.args
    while rest:
        expr, rest = partitionInst(rest, 1)
        exprs.append(expr)
    lms = [exprToLambda(e, fn_name, filename, args, fv) for e in reversed(exprs)]
    free_vars = set()
    for _, fvs in lms:
        if fvs:
            free_vars.update(fvs)
    resbc = bc.Bytecode(
        sum(
            map(get0, lms),
            start=(
                [bc.Instr("PUSH_NULL")]
                if sys.version_info.major == 3 and sys.version_info.minor >= 11
                else []
            ),
        )
    )
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        for cell in free_vars:
            resbc.insert(0, bc.Instr("MAKE_CELL", bc.CellVar(cell)))
    resbc.append(keynames)
    resbc.append(mk)
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        resbc.append(bc.Instr("LOAD_METHOD", "get"))
        for arg in args:
            resbc.append(bc.Instr("LOAD_FAST", arg))
        if len(args) > 1:
            resbc.append(bc.Instr("BUILD_TUPLE", arg=len(args)))
        resbc.extend(generateDefault(fn_name, filename, args))
        resbc.append(bc.Instr("PRECALL", 2))
        resbc.append(bc.Instr("CALL", 2))
        for arg in args:
            resbc.append(bc.Instr("LOAD_FAST", arg))
        resbc.append(bc.Instr("PRECALL", arg=len(args)))
        resbc.append(bc.Instr("CALL", arg=len(args)))
    else:
        resbc.append(bc.Instr("LOAD_METHOD", "get"))
        for arg in args:
            resbc.append(bc.Instr("LOAD_FAST", arg))
        resbc.append(bc.Instr("BUILD_TUPLE", arg=len(args)))
        resbc.extend(generateDefault(fn_name, filename, args))
        resbc.append(bc.Instr("CALL_METHOD", 2))
        for arg in args:
            resbc.append(bc.Instr("LOAD_FAST", arg))
        resbc.append(bc.Instr("CALL_FUNCTION", arg=len(args)))
    resbc.extend(case_inst.store.instrs)
    return resbc

def transVarMap(case_inst : Case, fn_name, filename, args, fv):
    body = case_inst.body.insts[:-1]
    mk = case_inst.body.insts[-1]
    parts = []
    rest = body
    hasDefault = False
    defaultExpr = []
    args = case_inst.head.args
    free_vars = set()
    while rest:
        part, rest = partitionInst(rest, 2)
        expr, pat = partitionInst(part, 1)
        if isWildcard(pat[0]) and len(pat) == 1:
            assert not hasDefault, f"Duplicated default cases in {fn_name}, line {pat[0].lineno} @ {filename}"
            hasDefault = True
            defaultExpr, fvs = exprToLambda(expr, fn_name, filename, args, fv)
            free_vars.update(fvs)
        else:
            vbind = {}
            vpat = []
            mkTpl = pat[-1]
            raw_pat_parts = pat[:-1] # fromRight([[], []], many(globalName | one(const(True)))(pat[:-1]))[0]
            pat_parts = []
            while raw_pat_parts:
                pat_part, raw_pat_parts = partitionInst(raw_pat_parts, 1)
                pat_parts.append(pat_part)
            for i, v in enumerate(reversed(pat_parts)):
                if len(v) == 1 and isVar(v[0]):
                    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
                        vbind[v[0].arg[1]] = args[i]
                        vpat.append(bc.Instr("LOAD_CONST", pat_core._v(v[0].arg[1])))
                    else:
                        vbind[v[0].arg] = args[i]
                        vpat.append(bc.Instr("LOAD_CONST", pat_core._v(v[0].arg)))
                else:
                    vpat.extend(v)
            vpat.append(mkTpl)
            lm, fvs = exprToLambda(expr, fn_name, filename, args, fv, vbind)
            parts.append((vpat, lm))
            free_vars.update(fvs)
    if not hasDefault:
        defaultExpr = generateDefault(fn_name, filename, args)
    defaultPat = []
    for _ in args:
        defaultPat.append(bc.Instr("LOAD_CONST", pat_core._v()))
    # if len(args) > 1:
    defaultPat.append(bc.Instr("BUILD_TUPLE", len(args)))
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        resbc = bc.Bytecode([bc.Instr("PUSH_NULL"), bc.Instr("PUSH_NULL"), bc.Instr("LOAD_CONST", pat_core.pytternd)])
    else:
        resbc = bc.Bytecode([bc.Instr("LOAD_CONST", pat_core.pytternd)])
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        for cell in free_vars:
            resbc.insert(0, bc.Instr("MAKE_CELL", bc.CellVar(cell)))
    for pat, expr in reversed(parts):
        resbc.extend(pat)
        resbc.extend(expr)
    resbc.extend(defaultPat)
    resbc.extend(defaultExpr)
    if not hasDefault:
        mk.arg += 1
    resbc.append(mk)
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        resbc.append(bc.Instr("PRECALL", 1))
        resbc.append(bc.Instr("CALL", 1))
    else:
        resbc.append(bc.Instr("CALL_FUNCTION", 1))
    for arg in args:
        resbc.append(bc.Instr("LOAD_FAST", arg))
    # if len(args) > 1:
    resbc.append(bc.Instr("BUILD_TUPLE", arg=len(args)))
    resbc.append(bc.Instr("BINARY_SUBSCR"))
    for arg in args:
        resbc.append(bc.Instr("LOAD_FAST", arg))
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        resbc.append(bc.Instr("PRECALL", arg=len(args)))
        resbc.append(bc.Instr("CALL", arg=len(args)))
    else:
        resbc.append(bc.Instr("CALL_FUNCTION", arg=len(args)))
    resbc.extend(case_inst.store.instrs)
    return resbc

@func
def handleCases(fn_name, filename, args, fv, instr):
    if not isinstance(instr, Case):
        return [instr]
    if mkConstMap(instr.body.insts[-1]):
        return transConstMap(instr, fn_name, filename, args, fv)
    elif mkMap(instr.body.insts[-1]):
        return transVarMap(instr, fn_name, filename, args, fv)

loadCase = and_(isLoadGlobal, __.arg ^ get1 == "case")

def extractGlobalName(i):
    if sys.version_info.major == 3 and sys.version_info.minor >= 11:
        return bc.Instr("LOAD_FAST", i.arg[1])
    return bc.Instr("LOAD_FAST", i.arg)

globalName = ptrans(one(isLoadGlobal), trans0(trans0(extractGlobalName)))

parseCaseHead = ptrans(
    one(loadCase) >> many1(one(isArg) | globalName) << callInst,
    trans0(mp1(__.arg) ^ (lambda x: [CaseHead(x)])),
) 

parseStoreName = (
    ptrans(one(popTop), trans0(trans0(const(StoreName([bc.Instr("POP_TOP")])))))
    | ptrans(one(storeFast), trans0(lambda x: [StoreName(x)]))
    | ptrans(one(unpack) + many1(one(storeFast)), trans0(lambda x: [StoreName(x)]))
)

parseCaseBody = ptrans(
    many1(neg(or_(mkConstMap, mkMap))) + one(or_(mkConstMap, mkMap)) << callInst,
    trans0((lambda x: [CaseBody(x)])),
)
parseCase = many(parseCaseHead + parseCaseBody + parseStoreName | one(const(True)))

mergeCase = many(ptrans(
    one(is_(CaseHead)) + one(is_(CaseBody)) + one(is_(StoreName)),
    trans0((lambda x: [Case(*x)]))
    ) | one(const(True)))

@do(Just)
def deco(b, fn_name, filename, args, freevars):
    transBc <- (rightToMaybe(parseCase(b)) | get0)
    mergeCase <- (rightToMaybe(mergeCase(transBc)) | get0)
    return bc.Bytecode(sum(mp1(handleCases(fn_name, filename, args, freevars), mergeCase), start = []))


def case(fn):
    rawbc = bc.Bytecode.from_code(fn.__code__)
    argcount = fn.__code__.co_argcount + (1 if isVarargFn(fn) else 0)
    args = fn.__code__.co_varnames[:argcount]

    # print( f"Generating pattern matching for function: {fn.__name__} at line: {rawbc.first_lineno} @ {rawbc.filename}")
    resbc = deco(rawbc, fn.__name__, rawbc.filename, args, rawbc.freevars)
    assert isJust(
        resbc
    ), f"Failed to generate pattern matching for function: {fn.__name__} at line: {rawbc.first_lineno} @ {rawbc.filename}"
    res = fromJust(resbc)
    # print(f"{res = }")
    cells = set(rawbc.cellvars)
    for inst in res:
        if isInstr(inst) and inst.name == "MAKE_CELL":
            cells.add(inst.arg.name)
    res.freevars.extend(rawbc.freevars)
    res.cellvars.extend(cells)
    res.argcount = rawbc.argcount
    res.argnames.extend(rawbc.argnames)
    res.name = rawbc.name
    res.filename = rawbc.filename
    res.flags = rawbc.flags
    res.update_flags()
    fn.__code__ = res.to_code()
    # print("=" * 20)
    # dis.dis(fn)
    # dis.show_code(fn)
    # print("=" * 20)
    return fn




from fpy.data.maybe import (
    Maybe,
    Just,
    Nothing,
    isJust,
    isNothing,
    fromJust,
)
from fpy.experimental.do import do

import unittest


def foo(x, y):
    return Just(x + y)


class TestDo(unittest.TestCase):
    def testSimpleJust(self):
        @do(Just)
        def test():
            x < -Just(1)
            return x

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 1)

    def testSimpleNothing(self):
        @do(Just)
        def test():
            x < -Just(1)
            Nothing()

        res = test()
        self.assertTrue(isNothing(res))

    def testLocal(self):
        @do(Just)
        def test():
            x = Just(1)
            y < -x
            return y

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 1)

    def testLocalNested(self):
        @do(Just)
        def test():
            x = 1
            y < -Just(2)
            z < -Just(x + y)
            return z

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 3)

    def testNested(self):
        @do(Just)
        def test():
            x < -Just(1)
            y < -Just(2)
            z < -foo(x, y)
            return z + z

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 6)

    def testTuple(self):
        @do(Just)
        def test():
            (a, b) < -Just((1, 2))
            return a + b

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 3)

    def testTupleNoParen(self):
        @do(Just)
        def test():
            a, b < -Just((1, 2))
            return a + b

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 3)

    def testComplex(self):
        @do(Just)
        def test():
            a, b <- Just((1, 2))
            c <- Just(a + b)
            d <- Just(c * b)
            return d

        res = test()
        self.assertTrue(isJust(res))
        self.assertEqual(fromJust(res), 6)



from fpy.data.cont import Cont, cont
from fpy.data.forgetful import forget, Under
from fpy.composable.function import func

import unittest


add = lambda a, b: a + b
mul = lambda a, b: a * b
div = lambda a, b: a / b


class TestCont(unittest.TestCase):
    def testCont(self):
        ac = cont(add)
        mc = cont(mul)
        dc = cont(div)
        res = dc(2, 2) >> mc(3) >> ac(1)
        self.assertEqual(Under(4), res & forget)
        self.assertEqual(Under(2), ac(1, 1) & forget)

    def testContWithFunc(self):
        ac = cont(func(add))
        mc = cont(func(mul))
        dc = cont(func(div))
        res = dc(2, 2) >> mc(3) >> ac(1)
        self.assertEqual(Under(4), res & forget)
        self.assertEqual(Under(2), ac(1, 1) & forget)



from __future__ import annotations

from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import TypeVar, Generic, Callable

from fpy.control.natural_transform import _NTrans
from fpy.composable.function import func

T = TypeVar("T")
F = TypeVar("F")
G = TypeVar("G")
A = TypeVar("A")
B = TypeVar("B")


class _Functor(Generic[T], metaclass=ABCMeta):
    @classmethod
    def fmap(cls, f: Callable[[A], B], a: _Functor[A]) -> _Functor[B]:
        assert isinstance(a, cls)
        return a.__fmap__(f)

    def __fmap__(self, f: Callable[[T], B]) -> _Functor[B]:
        raise NotImplementedError

    def __ntrans__(self, t: _NTrans[_Functor, T, G, B]) -> G[B]:
        """
        Natrual transformation:
        __ntrans__ :: f a -> (f a -> g b) -> g b
        """
        raise NotImplementedError

    def __and__(self, t: _NTrans[_Functor, T, G, B]) -> G[B]:
        return t.__trans__(self)

    def __or__(self, f):
        return self.__fmap__(f)


@dataclass
class Functor(_Functor[T], Generic[T]):
    val: T

    def __fmap__(self, f: Callable[[T], B]) -> _Functor[B]:
        return type(self)(f(self.val))

    def __ntrans__(self, t: Callable[[T], G[B]]) -> G[B]:
        return t(self.val)

@func
def fmap(o: F[A], f: Callable[[A], B]) -> F[B]:
    return o.__fmap__(f)



from abc import ABCMeta, abstractmethod, abstractclassmethod


class Composable(metaclass=ABCMeta):
    @abstractmethod
    def __compose__(self, other):
        raise NotImplementedError

    def __xor__(self, other):
        return self.__compose__(other)



from fpy.data.either import (
    Either,
    Left,
    Right,
    isLeft,
    isRight,
    fromLeft,
    fromRight,
    either,
    lefts,
    rights,
    partitionEithers,
    r2l,
    l2r,
)
from fpy.data.forgetful import forget, Under

import unittest


def even(n) -> Either[int, int]:
    if n % 2 == 0:
        return Right(n)
    return Left(n)


def add1(n):
    return n + 1


def sub1(n):
    return n - 1


class TestEither(unittest.TestCase):
    def testLeft(self):
        v = Left(3)
        self.assertTrue(isLeft(v))
        self.assertFalse(isRight(v))
        self.assertEqual(3, fromLeft(0, v))
        self.assertEqual(0, fromRight(0, v))
        self.assertEqual(4, either(add1, sub1, v))

    def testRight(self):
        v = Right(3)
        self.assertTrue(isRight(v))
        self.assertFalse(isLeft(v))
        self.assertEqual(3, fromRight(0, v))
        self.assertEqual(0, fromLeft(0, v))
        self.assertEqual(2, either(add1, sub1, v))

    def testEitherList(self):
        lst = [even(n) for n in range(8)]
        self.assertListEqual(lefts(lst), [1, 3, 5, 7])
        self.assertListEqual(rights(lst), [0, 2, 4, 6])
        ls, rs = partitionEithers(lst)
        self.assertListEqual(ls, [1, 3, 5, 7])
        self.assertListEqual(rs, [0, 2, 4, 6])

    def testNTransLeft(self):
        v = Left(3)
        vforgot = v & forget
        self.assertIsInstance(vforgot, Under)
        self.assertEqual(3, vforgot.under())
        vr = v & l2r
        self.assertTrue(isRight(vr))
        self.assertEqual(3, fromRight(0, vr))

    def testNTransRight(self):
        v = Right(3)
        vforgot = v & forget
        self.assertIsInstance(vforgot, Under)
        self.assertEqual(3, vforgot.under())
        vr = v & r2l
        self.assertTrue(isLeft(vr))
        self.assertEqual(3, fromLeft(0, vr))



from __future__ import annotations

from fpy.control.monad import _Monad
from fpy.composable.function import func, SignatureMismatchError

import bytecode as bc

from dataclasses import dataclass
from typing import TypeVar, Generic, Callable

import sys


T = TypeVar("T")
R = TypeVar("R")
S = TypeVar("S")


@dataclass
class Cont(_Monad[R], Generic[T, R]):
    f: Callable[[T], R]

    def __init__(self, f):
        self.f = func(f)

    def __bind__(self, f):
        return self.__class__.ret(lambda k: self.f(lambda *a: f(*a).__run__(k)))

    def __ntrans__(self, t):
        return self.f(t)

    def __call__(self, *a, **k):
        try:
            res = self.f(*a, **k)
            if isinstance(res, func) and res.f is self.f:
                return Cont(res)
            return Cont(lambda k: k(res))
        except SignatureMismatchError as e:
            raise e.e

    def __run__(self, k: Cont[[R], S]):
        return self.f(k)


def runCont(c, k):
    return c.__run__(k)


def to_cont(v):
    return Cont(lambda k: k(v))


def cont(f: Callable[[T], R]) -> Cont[T, R]:
    raw_bc = bc.Bytecode.from_code(f.__code__)

    print(raw_bc)
    res_bc = []
    for inst in raw_bc:
        if not isinstance(inst, bc.Instr):
            res_bc.append(inst)
            continue
        if inst.name != "RETURN_VALUE":
            res_bc.append(inst)
            continue

        res_bc.append(bc.Instr("LOAD_CONST", to_cont, lineno=inst.lineno))
        if sys.version_info.major == 3 and sys.version_info.minor >= 11:
            res_bc.append(bc.Instr("PUSH_NULL", lineno=inst.lineno))
            res_bc.append(bc.Instr("SWAP", 3, lineno=inst.lineno))
            res_bc.append(bc.Instr("PRECALL", 1))
            res_bc.append(bc.Instr("CALL", 1))
        else:
            res_bc.append(bc.Instr("ROT_TWO", lineno=inst.lineno))
            res_bc.append(bc.Instr("CALL_FUNCTION", 1, lineno=inst.lineno))
        res_bc.append(inst)
    print(res_bc)
    bc_obj = bc.Bytecode(res_bc)
    bc_obj.freevars = f.__code__.co_freevars
    bc_obj.cellvars = f.__code__.co_cellvars
    bc_obj.argcount = f.__code__.co_argcount
    bc_obj.name = f.__name__
    f.__code__ = bc_obj.to_code()
    return func(f)
